/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

const EXPORTED_SYMBOLS = ["NewWindow",
                          "Profile",         // error.js
                          "WindowProperties" // about:multifox
                         ];

Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");
Components.utils.import("resource://multifox-modules/new-window.js");

util.log("===>LOADING main.js");


const NewWindow = {
  newId: function(win) {
    var id;
    if (this._shouldBeDefault(win)) {
      id = Profile.DefaultIdentity;
    } else {
      id = Profile.lowerAvailableId(win);
    }
    util.log("newIdentity " + id);
    Profile.defineIdentity(win, id);
  },

  inheritId: function(newWin) {
    util.log("inheritId");
    var id;
    if (this._shouldBeDefault(newWin)) {
      id = Profile.DefaultIdentity;
    } else {
      //window.open()/fxstarting ==> opener=null
      var prevWin = newWin.opener;
      if (prevWin) {
        id = Profile.getIdentity(prevWin);
      } else {
        util.log("inheritId prevWin=" + prevWin);
        id = Profile.UnknownIdentity;
      }
    }
    Profile.defineIdentity(newWin, id);
    util.log("/inheritId " + id);
  },

  applyRestore: function(win) {
    // restore: window is first configured by NewWindow.inheritId
    util.log("applyRestore");

    var stringId = Cc["@mozilla.org/browser/sessionstore;1"]
                    .getService(Ci.nsISessionStore)
                    .getWindowValue(win, "multifox-dom-identity-id");
    var id = Profile.toInt(stringId);
    if (id < Profile.DefaultIdentity) { //UnknownIdentity?
      id = Profile.DefaultIdentity;
    }
    Profile.defineIdentity(win, id);
  },

  _shouldBeDefault: function(win) {
    // popup opened by an extension (like GTB)
    //var chromeHidden = win.document.documentElement.getAttribute("chromehidden");
    //return chromeHidden.indexOf("location") > -1;
    return false;
  }

};


const Profile = {
  UnknownIdentity: 0,
  DefaultIdentity: 1,
  MaxIdentity:     999999999999999,

  defineIdentity: function(win, id) {
    util.log("defineIdentity " + id);
    if (id > Profile.MaxIdentity) {
      util.log("id > max " + id);
      id = Profile.MaxIdentity;
    }
    if (id < Profile.UnknownIdentity) {
      util.log("id < UnknownIdentity " + id);
      id = Profile.UnknownIdentity;
    }
    var current = Profile.getIdentity(win);
    if (current === id) {
      util.log("defineIdentity NOP");
      return id;
    }
    if (current !== Profile.DefaultIdentity) {
      BrowserWindow.unregister(win);
    }

    this._save(win, id);
    BrowserWindow.register(win);
    updateUI(win);
    return id;
  },


  getIdentity: function(chromeWin) {
    var tabbrowser = chromeWin.getBrowser();
    if (tabbrowser === null) {
      util.log("getIdentity=DefaultIdentity, tabbrowser=null");
      return Profile.DefaultIdentity;
    }
    var id = tabbrowser.getUserData("multifox-dom-identity-id");
    return this.toInt(id);
  },

  _save: function(win, id) {
    util.log("save " + id);
    win.getBrowser().setUserData("multifox-dom-identity-id", id === Profile.DefaultIdentity ? null : id, null);
    new SaveToSessionStore(win.document);
  },

  activeIdentities: function(ignoreWin) {
    var winEnum = util2.browserWindowsEnum();
    var arr = [];
    while (winEnum.hasMoreElements()) {
      var win = winEnum.getNext();
      if (ignoreWin !== win) {
        var id = Profile.getIdentity(win);
        if (arr.indexOf(id) === -1) {
          arr.push(id);
        }
      }
    }
    return arr;
  },

  lowerAvailableId: function(ignoreWin) {
    var arr = this.activeIdentities(ignoreWin); //ignore win -- it doesn't have a session id yet
    var id = Profile.DefaultIdentity;
    while (arr.indexOf(id) > -1) {
      id++;
    }
    return id; // id is available
  },

  toInt: function(str) {
    var rv = parseInt(str, 10);
    return isNaN(rv) ? Profile.DefaultIdentity : rv;
  },

  toString: function(id) {
    switch (id) {
      //case Profile.UnknownIdentity:
      //  return "\u221e"; // ∞
      default:
        return id.toString();
    }
  }

};


function SaveToSessionStore(doc) {
  this._doc = doc;
  this._timer = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
  this._timer.init(this, 1300, Ci.nsITimer.TYPE_ONE_SHOT);
}

SaveToSessionStore.prototype = {

  QueryInterface: XPCOMUtils.generateQI([Ci.nsIObserver, Ci.nsISupportsWeakReference]),

  observe: function(aSubject, aTopic, aData) {
    util.log("SaveToSessionStore timer!");
    var doc = this._doc;
    if (doc.defaultView === null) {
      util.log("_syncToSessionStore window closed");
      return;
    }

    var ss = Cc["@mozilla.org/browser/sessionstore;1"].getService(Ci.nsISessionStore);
    var val = Profile.getIdentity(doc.defaultView);

    try {
      // overwrite any previous value if called twice
      ss.setWindowValue(doc.defaultView, "multifox-dom-identity-id", val);
    } catch (ex) {
      // keep trying
      util2.logEx("SaveToSessionStore FAIL", val, doc, doc.defaultView, doc.defaultView.state, ex);
      this._timer.init(this, 700, Ci.nsITimer.TYPE_ONE_SHOT);
      return;
    }

    if (val <= Profile.DefaultIdentity) { // UnknownIdentity OR DefaultIdentity
      ss.deleteWindowValue(doc.defaultView, "multifox-dom-identity-id");
    }

    util.log("_syncToSessionStore OK=" + Profile.getIdentity(doc.defaultView));
  }
};
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

const BrowserWindow = {

  register: function(win) {
    var profileId = Profile.getIdentity(win);

    if (profileId === Profile.DefaultIdentity) {
      util.log("BrowserWindow.register NOP");
      return;
    }

    util.log("BrowserWindow.register " + profileId);

    if (util.networkListeners.active === false) {
      // first multifox window!
      multifoxStartup();
    }

    DocStartScriptInjection.register(win.getBrowser());

    win.addEventListener("MultifoxContentEvent_FromContent", onContentEvent, false, true);

    // some MultifoxContentEvent_* listeners are not called when
    // there are "unload" listeners with useCapture=true. o_O
    // But they are called if event listener is an anonymous function.
    win.addEventListener("unload", onUnloadChromeWindow, false);


    // restore icon after toolbar customization
    var toolbox = win.document.getElementById("navigator-toolbox");
    toolbox.addEventListener("DOMNodeInserted", customizeToolbar, false);
  },


  // should keep id for session restore?
  unregister: function(win) {
    var idw = Profile.getIdentity(win);
    util.log("BrowserWindow.unregister " + idw);

    if (idw === Profile.DefaultIdentity) {
      // nothing to unregister
      return;
    }


    DocStartScriptInjection.unregister(win.getBrowser());
    win.removeEventListener("MultifoxContentEvent_FromContent", onContentEvent, false);

    var sessions = Profile.activeIdentities(win);
    var toolbox = win.document.getElementById("navigator-toolbox");
    toolbox.removeEventListener("DOMNodeInserted", customizeToolbar, false);

    var onlyDefault = (sessions.length === 1) && (sessions[0] === Profile.DefaultIdentity);
    if (onlyDefault) {
      // no more multifox windows
      multifoxShutdown();
    }
    win.removeEventListener("unload", onUnloadChromeWindow, false);
  }
};


function onUnloadChromeWindow(evt) {
  var win = evt.currentTarget;
  BrowserWindow.unregister(win);
}


function customizeToolbar(evt) {
  var node = evt.target;
  if ((node.id === "urlbar-container") && (node.parentNode.tagName === "toolbar")) {
    updateUI(node.ownerDocument.defaultView);
  }
}


function onContentEvent(evt) {
  evt.stopPropagation();

  var obj = JSON.parse(evt.data);
  var contentDoc = evt.target;
  var rv;

  switch (obj.from) {
    case "cookie":
      rv = documentCookie(obj, contentDoc);
      break;
    case "localStorage":
      rv = windowLocalStorage(obj, contentDoc);
      break;
    case "error":
      Components.utils.import("resource://multifox-modules/error.js");
      showError(contentDoc.defaultView, obj.cmd);
      break;
    default:
      throw obj.from;
  }

  if (rv === undefined) {
    // no response
    return;
  }

  // send data to content
  var evt2 = contentDoc.createEvent("MessageEvent");
  evt2.initMessageEvent("MultifoxContentEvent_FromContent_Response", false, false, rv, null, null, null);
  var success = contentDoc.dispatchEvent(evt2);
}


function multifoxStartup() {
  util.log("multifoxStartup");
  Cookies.start();
  util.networkListeners.enable(httpListeners.request, httpListeners.response);
  DocStartScriptInjection.init("chrome://multifox/content/content-injection.js");
  WatchObjectsStatus.start();
  //toggleDefaultWindowUI(true);
  util.log("/multifoxStartup");
}

function multifoxShutdown() {
  util.log("multifoxShutdown");
  util.networkListeners.disable();
  DocStartScriptInjection.shutdown();
  Cookies.stop();
  WatchObjectsStatus.stop();
  //toggleDefaultWindowUI(false);
}


const WatchObjectsStatus = {
  _timer: null,

  start: function() {
    this._timer = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
    this._schedule();
  },

  stop: function() {
    this._timer.cancel();
    this._timer = null;
  },

  _schedule: function() {
    this._timer.init(this, 90000, Ci.nsITimer.TYPE_ONE_SHOT);
  },

  QueryInterface: XPCOMUtils.generateQI([Ci.nsIObserver, Ci.nsISupportsWeakReference]),

  observe: function(aSubject, aTopic, aData) {
    var buf = "watchObjectsStatus\n";

    buf += "_pendingFrames.length="+ PendingFrames._pendingFrames.length + " jsd=" + (Jsd._service ? Jsd._service.isOn : "null");

    var qty = PendingFrames._pendingFrames.length;
    for (var idx = qty - 1; idx >= 0; idx--) {
      buf += "\n" + idx + " " + PendingFrames._pendingFrames[idx].src;
    }

    util.log(buf);
    this._schedule();
  }
};
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

function getIconNode(doc) {
  if (!doc) {
    util2.throwStack.go("getIconNode "+doc);
  }
  return doc.getElementById("multifox-icon");
}


function removeUI(doc) {
  util.log("removeUI");
  var icon = getIconNode(doc);
  if (icon) {
    icon.parentNode.removeChild(icon);
  }
}


function updateUI(win) {
  var profileId = Profile.getIdentity(win);
  util.log("updateUI " + profileId);

  var doc = win.document;
  if (profileId <= Profile.DefaultIdentity) {
    removeUI(doc);
    return;
  }

  var icon = getIconNode(doc);
  if (icon) {
    util.log("updateUI=" + profileId + " current=" + Profile.getIdentity(win));
    icon.querySelector("label").setAttribute("value", Profile.toString(profileId));
    return;
  }


  // <hbox align="center" id="multifox-icon">
  //   <box>
  //     <hbox align="center">
  //       <label value="100"/>


  var ref = doc.getElementById("urlbar-icons");
  var iconContainer = ref.appendChild(doc.createElement("hbox"));
  iconContainer.setAttribute("hidden", "true");
  iconContainer.setAttribute("id", "multifox-icon");


  if (profileId !== Profile.UnknownIdentity) {
    win.setTimeout(initIconCore, 250, iconContainer, profileId);
  }

}


function initIconCore(iconContainer, profileId) {
  var doc = iconContainer.ownerDocument;
  if (iconContainer !== getIconNode(doc)) {
    util.log("initIconCore != " + profileId + iconContainer + getIconNode(doc));
    return;
  }


  iconContainer.setAttribute("align", "center");
  iconContainer.style.margin = "0px 2px 0px 1px";

  var icon = iconContainer.appendChild(doc.createElement("box"));
  icon.setAttribute("tooltiptext", "Multifox");


  var styleIcon = icon.style;
  styleIcon.padding = "0px";
  styleIcon.margin = "0px";
  styleIcon.background = "transparent repeat-x";
  styleIcon.MozBorderRadius = "2px";
  styleIcon.minHeight = "16px";


  var os = Cc["@mozilla.org/xre/app-info;1"].getService(Ci.nsIXULRuntime).OS;
  switch (os) {
    case "Darwin":
      styleIcon.backgroundImage = "url(chrome://multifox/content/icon-osx.png)";
      styleIcon.backgroundColor = "#416ea5";
      break;
    case "WINNT":
      styleIcon.backgroundImage = "url(chrome://multifox/content/icon.png)";
      styleIcon.backgroundColor = "#4ab0f6";
      break;
    default:
      styleIcon.backgroundImage = "url(chrome://multifox/content/icon-linux.png)";
      styleIcon.backgroundColor = "#6184c3";
      break;
  }



  var labelContainer = icon.appendChild(doc.createElement("hbox"));
  labelContainer.setAttribute("align", "center");


  var lb = labelContainer.appendChild(doc.createElement("label"));
  lb.setAttribute("value", Profile.toString(profileId));
  var styleLabel = lb.style;

  styleLabel.color = "white";
  styleLabel.fontWeight = "bold";
  styleLabel.fontStyle = "normal";
  styleLabel.textRendering = "optimizelegibility";
  styleLabel.textShadow = "1px 1px 1px black";
  styleLabel.padding = "0px 9px";
  styleLabel.margin = "-1px 0px";

  initIconNormal(icon);
  iconContainer.removeAttribute("hidden");
}


function initIconNormal(icon) {
  icon.ownerDocument.defaultView.setTimeout(function() {
    // workaround
    // otherwise, mousedown on icon will close and open popup, even with stopPropagation
    icon.addEventListener("mousedown", openMultifoxPopup, false);
  }, 0);
  icon.addEventListener("mouseover", onIconHover, false);
  icon.addEventListener("mouseout", onIconHover, false);
  icon.style.backgroundPosition = "0pt 0pt";
  icon.querySelector("label").style.textShadow = "1px 1px 1px black";
}


function initIconPressed(icon) { // openpopup
  icon.removeEventListener("mousedown", openMultifoxPopup, false);
  icon.removeEventListener("mouseover", onIconHover, false);
  icon.removeEventListener("mouseout", onIconHover, false);
  icon.style.backgroundPosition = "0pt -64px";
  icon.querySelector("label").style.textShadow = "1px 1px 1px black, 0 1px 5px white, 3px 0 5px white, 0 -1px 5px white, -3px 0 5px white";
}


function onIconHover(evt) {
  var icon = getIconNode(evt.target.ownerDocument).firstChild;

  switch (evt.type) {
    case "mouseover":
      icon.style.backgroundPosition = "0pt -32px";
      icon.querySelector("label").style.textShadow = "1px 1px 1px black, 0 1px 5px white, 3px 0 5px white, 0 -1px 5px white, -3px 0 5px white";
      break;
    case "mouseout":
      icon.style.backgroundPosition = "0pt 0pt";
      icon.querySelector("label").style.textShadow = "1px 1px 1px black";
      break;
  }
}


/*
// show/hide [1] "icon"
function toggleDefaultWindowUI(show) {
  var winEnum = util2.browserWindowsEnum();
  while (winEnum.hasMoreElements()) {
    var doc = winEnum.getNext().document;
    if (Profile.getIdentity(doc) === Profile.DefaultIdentity) {
      if (show) {
        updateUI(doc, Profile.DefaultIdentity);
      } else {
        removeUI(doc);
      }
    }
  }
}
*/


function openMultifoxPopup(evt) {
  Components.utils.import("resource://multifox-modules/popup.js");
  var icon = evt.currentTarget; //or this

  var panel = createMultifoxPopup(icon, Profile);
  initIconPressed(icon);

  panel.addEventListener("popuphidden", function(evt) {
    initIconNormal(icon);
  }, false);

  panel.openPopup(icon, "after_end", 0, 1);
}
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// frame/iframe create/starts loading
//   -current page unloads (about:blank for new elements)
//   -pagehide is triggered
//     -frame node added to _pendingFrames
//     -jsd on
//   -new page starts loading
//   -Jsd.topLevelHook.onCall triggered
//     -loop _pendingFrames
//       -onDocStart

// Most of this code can be removed as soon bug 342715 is fixed.


const DocStartScriptInjection = {

  init: function(pathSrc) {
    BootSource.init(pathSrc);
    PendingFrames.init();
  },

  shutdown: function() {
    BootSource.shutdown();
    PendingFrames.shutdown();
  },

  register: function(elem) {
    util.log("DocStartScriptInjection.register " + elem.tagName);
    switch (elem.tagName) {
      case "browser":
        elem.addProgressListener(this._browserElement, Ci.nsIWebProgress.NOTIFY_ALL);
        break;
      case "tabbrowser":
        elem.addTabsProgressListener(this._tabbrowserElement);
        break;
      default:
        throw elem.tagName;
    }
    elem.addEventListener("DOMContentLoaded", this._onDOMLoaded, false);
    elem.addEventListener("pagehide", this._onPageHide, false);
  },

  unregister: function(elem) {
    util.log("DocStartScriptInjection.unregister " + elem.tagName);
    switch (elem.tagName) {
      case "browser":
        elem.removeProgressListener(this._browserElement);
        break;
      case "tabbrowser":
        elem.removeTabsProgressListener(this._tabbrowserElement);
        break;
      default:
        throw elem.tagName;
    }
    elem.removeEventListener("DOMContentLoaded", this._onDOMLoaded, false);
    elem.removeEventListener("pagehide", this._onPageHide, false);
  },

  _tabbrowserElement: {
    QueryInterface: XPCOMUtils.generateQI([Ci.nsIWebProgressListener, Ci.nsISupportsWeakReference]),
    onStateChange: function() {},
    onSecurityChange: function() {},
    onProgressChange: function() {},
    onStatusChange: function() {},
    onLinkIconAvailable: function() {},
    onLocationChange: function(aBrowser, aWebProgress, aRequest, aLocation) {
      // for frames: it fires only if DOMWindow is different from top window
      BootSource.onDocStart(aWebProgress.DOMWindow);
    }
  },

  _browserElement: {
    QueryInterface: XPCOMUtils.generateQI([Ci.nsIWebProgressListener, Ci.nsISupportsWeakReference]),
    onStateChange: function() {},
    onSecurityChange: function() {},
    onProgressChange: function() {},
    onStatusChange: function() {},
    onLinkIconAvailable: function() {},
    onLocationChange: function(aWebProgress, aRequest, aLocation) {
      BootSource.onDocStart(aWebProgress.DOMWindow);
    }
  },

  _onPageHide: function(evt) {
    // new frames (about:blank) or new url
    var doc = evt.target;
    var frm = doc.defaultView.frameElement;
    if (frm !== null) { // top-level window?
      PendingFrames.addPendingFrame(frm);
    }
  },

  _onDOMLoaded: function(evt) {
    var doc = evt.target;
    var frm = doc.defaultView.frameElement;
    if (frm !== null) { // top-level window?
      PendingFrames.domLoaded(frm);
    }
  }

};


const BootSource = {
  init: function(pathSrc) {
    this._path = pathSrc;
    this._load(true);
  },

  shutdown: function() {
    delete this._path;
    delete this._src;
  },

  _load: function(async) {
    var xhr = Cc["@mozilla.org/xmlextras/xmlhttprequest;1"].createInstance(Ci.nsIXMLHttpRequest);
    xhr.onload = function() {
      BootSource._src = xhr.responseText;
      BootSource._path = null;
    };
    xhr.open("GET", this._path, async);
    xhr.overrideMimeType("text/plain");
    xhr.send(null);
  },

  onDocStart: function(win) {
    if (win._multifox_hooked) {
      win._multifox_hooked++;
      return;
    }

    if (this._src === null) {
      this._load(false);
    }

    var wrappedWin = new XPCNativeWrapper(win);
    var sandbox = new Components.utils.Sandbox(wrappedWin);
    sandbox.window = wrappedWin;
    //sandbox.document = wrappedWin.document;
    //sandbox.__proto__ = wrappedWin;

    try {
      // TODO http/s only
      Components.utils.evalInSandbox(this._src, sandbox);
      win._multifox_hooked = 1;
    } catch (ex) {
      switch (wrappedWin.document.documentURIObject.scheme) {
        case "http":
        case "https":
          Components.utils.import("resource://multifox-modules/error.js");
          showError(wrappedWin, "sandbox");
          break;
      }
      util.log("[sandbox] " + win.document.location + "\n====8<-----\n" + ex);
    }

  }

};


const PendingFrames = {
  init: function() {
    this._pendingFrames = [];
    Jsd.disable();
  },

  shutdown: function() {
    Jsd.disable();
    delete this._pendingFrames;
  },

  addPendingFrame: function(frm) {
    if (this._pendingFrames.indexOf(frm) > -1) {
      util.log("pagehide event: frame already exists in _pendingFrames. " + frm.src + this._pendingFrames.indexOf(frm));
      return;
    }

    // when a document unloads, "pagehide" is triggered. This frame is useless. How to detect it?
    this._pendingFrames.push(frm);
    frm.contentWindow._multifox_pagehide = 1;
    ScheduleGc.start();
    Jsd.enable();

    // frm.tagName="object","iframe"
    //util.log("pagehide event: new frame added! " + frm.tagName  + " src=" + frm.src.substr(0, 80) + " doc=" + frm.contentDocument.location.href + " innerHTML=[" + frm.innerHTML + "]");
  },


  _removePendingFrame: function(idx) {
    this._pendingFrames.splice(idx, 1);
    if (this._pendingFrames.length === 0) {
      Jsd.disable(); // disable jsd asap
    }
  },

  tryToDispose: function() {
    for (var idx = this._pendingFrames.length - 1; idx > -1; idx--) {

      if (this._frameExists(this._pendingFrames[idx]) === false) {
        this._removePendingFrame(idx);
        continue;
      }

      var frameWin = this._pendingFrames[idx].contentWindow;

      if (!frameWin) {
        this._removePendingFrame(idx);
        util.log("tryToDispose frameWin empty "+frameWin);
        continue;
      }

      if (frameWin._multifox_pagehide) {
        util.log("tryToDispose frameWin._multifox_pagehide "+frameWin.document.location + frameWin._multifox_hooked + this._frameExists(this._pendingFrames[idx]));
        continue;
      }


      if (frameWin._multifox_hooked) {
        this._removePendingFrame(idx);
        util.log("tryToDispose _multifox_hooked="+frameWin);
      }
    }
    return this._pendingFrames.length > 0;
  },

  domLoaded: function(frm) {
    var idx = this._pendingFrames.indexOf(frm);
    if (idx > -1) {
      // frame without a <script>
      this._removePendingFrame(idx);
      BootSource.onDocStart(frm.contentWindow);
    }
  },

  onJavaScriptFound: function() {
    // we don't know which document Jsd.topLevelHook is referring to
    // we have to call BootSource.onDocStart for all pending frames
    var pending = this._pendingFrames;
    for (var idx = pending.length - 1; idx >= 0; idx--) {
      var frameWin = pending[idx] ? pending[idx].contentWindow : null;
      if (!frameWin) {
        // ???
        this._removePendingFrame(idx);
        util.log("onJavaScriptFound frameWin=" + frameWin + " pending[idx]=" + pending[idx] + " idx=" + idx);
        continue;
      }
      if (frameWin._multifox_pagehide) {
        // page not loaded yet
        continue;
      }

      this._removePendingFrame(idx);
      if (!frameWin._multifox_hooked) {
        BootSource.onDocStart(frameWin);
      }
    }
  },

  _frameExists: function(frm) {
    var winFrame = frm.contentWindow;

    if (!winFrame) { // null most of the time, sometimes it is undefined
      // win closed?
      util.log("frame does not exist! (win closed/frame removed?) src=" + frm.src);
      return false;
    }

    var myFrame = frm;
    var rootNotFound = true;
    do {
      var parentFrame = null;
      if (myFrame.ownerDocument.defaultView) {
        parentFrame = myFrame.ownerDocument.defaultView.frameElement;
      }

      if (parentFrame === null) {
        var parentFrame = new WindowProperties(winFrame).browser;
        if (parentFrame === null) {
          util.log("frame does not exist! (it does not have a parent)");
          return false;
        }
        rootNotFound = false;
      }
      var pos = parentFrame.contentDocument.documentElement.compareDocumentPosition(myFrame);
      if ((pos & 1) === 1) { // DOCUMENT_POSITION_DISCONNECTED=1
        util.log("frame does not exist! (frame removed?)");
        return false;
      }
      myFrame = parentFrame;
    } while (rootNotFound);

    return true;
  }

};


const ScheduleGc = {
  _timer: null,

  start: function() {
    if (this._timer === null) {
      this._timer = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
      this._timer.init(this, 200, Ci.nsITimer.TYPE_ONE_SHOT);
    }
  },

  QueryInterface: XPCOMUtils.generateQI([Ci.nsIObserver, Ci.nsISupportsWeakReference]),

  observe: function(aSubject, aTopic, aData) {
    this._timer = null;
    if (PendingFrames.tryToDispose()) {
      this.start();
      util.log("ScheduleGc timer renewed");
    }
  }
};


const Jsd = {
  enable: function() {
    if (this._service !== null) {
      return;
    }
    //util.log("jsd ON");
    var i = Ci.jsdIDebuggerService;
    var service = Cc["@mozilla.org/js/jsd/debugger-service;1"].getService(i);
    service.topLevelHook = this;
    service.on();
    service.flags |= i.DISABLE_OBJECT_TRACE;
    this._service = service;
  },

  disable: function() {
    util.log("jsd OFF");
    if (this._service) {
      this._service.off();
    }
    this._service = null;
  },

  _service: null,

  QueryInterface: XPCOMUtils.generateQI([Ci.jsdICallHook]),
  onCall: function(frame, type) {
    if (type === Ci.jsdICallHook.TYPE_TOPLEVEL_START) {
      PendingFrames.onJavaScriptFound();
    }
  }

};
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

const httpListeners = {
  request: {
    QueryInterface: XPCOMUtils.generateQI([Ci.nsIObserver]),

    // nsIObserver
    observe: function(aSubject, aTopic, aData) {
      var httpChannel = aSubject.QueryInterface(Ci.nsIHttpChannel);
      var info = new WindowProperties(getChannelWindow(httpChannel));
      switch (info.identity) {
        case Profile.DefaultIdentity:
        case Profile.UnknownIdentity: // favicon, updates
          return;
      }

      var myHeaders = HttpHeaders.fromRequest(httpChannel);
      if (myHeaders["authorization"] !== null) {
        Components.utils.import("resource://multifox-modules/error.js");
        showError(info.contentWindow, "authorization");
        return;
      }

      var cook = Cookies.getCookie(false, httpChannel.URI, info.identity);
      httpChannel.setRequestHeader("Cookie", cook, false);
    }
  },

  response: {
    QueryInterface: XPCOMUtils.generateQI([Ci.nsIObserver]),

    // nsIObserver
    observe: function(aSubject, aTopic, aData) {
      var httpChannel = aSubject.QueryInterface(Ci.nsIHttpChannel);
      var info = new WindowProperties(getChannelWindow(httpChannel));
      switch (info.identity) {
        case Profile.DefaultIdentity:
        case Profile.UnknownIdentity:
          return;
      }


      var myHeaders = HttpHeaders.fromResponse(httpChannel);
      if (myHeaders["www-authenticate"] !== null) {
        Components.utils.import("resource://multifox-modules/error.js");
        showError(info.contentWindow, "www-authenticate");
        return;
      }

      var setCookies = myHeaders["set-cookie"];
      if (setCookies === null) {
        return;
      }

      // server sent "Set-Cookie"
      httpChannel.setResponseHeader("Set-Cookie", null, false);
      Cookies.setCookie(info.identity, httpChannel.URI, setCookies, false);
    }
  }
};


const HttpHeaders = {
  visitLoop: {
    values: null,
    visitHeader: function(name, value) {
      var n = name.toLowerCase();
      if (n in this.values) {
        this.values[n] = value;
      }
    }
  },

  fromRequest: function(request) {
    var nameValues = {
      //"cookie": null, //for debug only
      "authorization": null
    }
    this.visitLoop.values = nameValues;
    request.visitRequestHeaders(this.visitLoop);
    return nameValues;
  },

  fromResponse: function(response) {
    var nameValues = {
      "set-cookie": null,
      "www-authenticate": null
    }
    this.visitLoop.values = nameValues;
    response.visitResponseHeaders(this.visitLoop);
    return nameValues;
  }
};


function getChannelWindow(channel) {
  if (channel.notificationCallbacks) {
    try {
      return channel
              .notificationCallbacks
              .getInterface(Ci.nsILoadContext)
              .associatedWindow;
    } catch (ex) {
      //util2.logEx("channel.notificationCallbacks ", channel.notificationCallbacks, channel.URI.spec, ex);
    }
  }

  if (channel.loadGroup && channel.loadGroup.notificationCallbacks) {
    try {
      return channel
              .loadGroup
              .notificationCallbacks
              .getInterface(Ci.nsILoadContext)
              .associatedWindow;
    } catch (ex) {
      util2.logEx("channel.loadGroup", channel.loadGroup, channel.URI.spec, ex);
    }
  }

  //var isChrome = context.associatedWindow instanceof Ci.nsIDOMChromeWindow;
  //return context.isContent ? context.associatedWindow : null;
  //util.log("LOAD CONTEXT FAIL " + channel.URI.spec);
  return null; // e.g. <link rel=prefetch> <link rel=next> ...
}
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2010
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */


function WindowProperties(contentWin) {
  if (contentWin === null) {
    return;
  }
  this._content = contentWin;
  this._browser = this._findBrowser(contentWin);
  if (this._browser) {
    // browser.xul
    this._chrome = this._browser.ownerDocument.defaultView;
  } else {
    this._chrome = this._findChromeWindow(contentWin);
  }
}

WindowProperties.prototype = {
  _content: null,
  _browser: null,
  _chrome: null,
  _profile: -1,

  _findChromeWindow: function(contentWin) {
    if (!contentWin) {
      util2.logEx("_findChromeWindow contentWin", contentWin);
    }

    var docShellTreeItem = contentWin
                            .QueryInterface(Ci.nsIInterfaceRequestor)
                            .getInterface(Ci.nsIWebNavigation)
                            .QueryInterface(Ci.nsIDocShellTreeItem);

    var win = docShellTreeItem
                .rootTreeItem
                .QueryInterface(Ci.nsIInterfaceRequestor)
                .getInterface(Ci.nsIDOMWindow);

    if (win instanceof Ci.nsIDOMChromeWindow) {
      // wrappedJSObject allows access to gBrowser etc
      // it looks like wrappedJSObject=undefined for chrome URLs
      return win.wrappedJSObject ? win.wrappedJSObject : win;
    }

    // it doesn't work sometimes (docShellTreeItem.treeOwner is null)
    util2.logEx("_findChromeWindow", contentWin.document.location, win, this._findBrowser2(contentWin));
    return null;
  },

  _findBrowser: function(contentWin) {
    var chromeWindow = this._findChromeWindow(contentWin);
    if (chromeWindow === null) {
      util.log("_findBrowser - chromeWindow=null. win=" + contentWin.document.location);
      return this._findBrowser2(contentWin);
    }

    var type = chromeWindow.document.documentElement.getAttribute("windowtype");
    if (type !== "navigator:browser") {
      return null;
    }

    var tabbrowser = chromeWindow.getBrowser();
    var idx = tabbrowser.getBrowserIndexForDocument(contentWin.top.document);
    return idx > -1 ? tabbrowser.browsers[idx] : null;
  },


  _findBrowser2: function(contentWin) {
    var topDoc = contentWin.top.document;
    var winEnum = util2.browserWindowsEnum();
    while (winEnum.hasMoreElements()) {
      var tabbrowser = winEnum.getNext().getBrowser();
      var idx = tabbrowser.getBrowserIndexForDocument(topDoc);
      if (idx > -1) {
        return tabbrowser.browsers[idx];
      }
    }
    return null;
  },

  get contentWindow() {
    return this._content;
  },

  get chromeWindow() {
    return this._chrome;
  },

  get browser() {
    return this._browser;
  },

  get identity() {
    if (this._profile === -1) {
      this._profile = this._getIdentity();
    }
    return this._profile;
  },

  _getIdentity: function() {
    if (this.contentWindow === null) {
      return Profile.UnknownIdentity;
    }

    if (this.chromeWindow === null) {
      return Profile.UnknownIdentity;
    }

    if (this.browser === null) {
      // source-view?
      return this._getIdentityFromOpenerChrome();
    }


    var profileId = Profile.getIdentity(this.chromeWindow);
    if (profileId !== Profile.UnknownIdentity) {
      return profileId;
    }

    // popup via js/window.open
    return this._getIdentityFromOpenerContent();
  },

  _getIdentityFromOpenerContent: function() {
    if (this.contentWindow.opener) {
      var browserOpener = this._findBrowser(this.contentWindow.opener);
      if (browserOpener) {
        var chromeOpener = browserOpener.ownerDocument.defaultView;
        var profileId = Profile.getIdentity(chromeOpener);
        if (profileId > Profile.UnknownIdentity) {
          return Profile.defineIdentity(this.chromeWindow, profileId);
        }
      }
    }

    util.log("request [" + profileId + "] id.opener="+ profileId);
    return Profile.UnknownIdentity;
  },

  _getIdentityFromOpenerChrome: function() {
    var tabbrowser = null;
    var type = this.chromeWindow.document.documentElement.getAttribute("windowtype");
    if (type === "navigator:view-source") {
      var winOpener = this.chromeWindow.opener;
      if (winOpener) {
        if (winOpener.document.documentElement.getAttribute("windowtype") === "navigator:browser") {
          tabbrowser = winOpener.getBrowser();
        }
      }
    }

    return tabbrowser !== null ? Profile.getIdentity(tabbrowser.ownerDocument.defaultView)
                               : Profile.UnknownIdentity; // favicon, ...
  }

};
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */


function documentCookie(obj, contentDoc) {
  switch (obj.cmd) {
    case "set":
      documentCookieSetter(obj, contentDoc);
      return undefined;
    case "get":
      return documentCookieGetter(obj, contentDoc);
    default:
      throw obj.cmd;
  }
}


function documentCookieGetter(obj, contentDoc) {
  var profileId = new WindowProperties(contentDoc.defaultView).identity;

  switch (profileId) {
    case Profile.UnknownIdentity:
      return;
    case Profile.DefaultIdentity:
      util2.throwStack.go("documentCookieGetter " + profileId);
      return;
  }

  var uri = util2.stringToUri(contentDoc.location.href);
  var cookie2 = Cookies.getCookie(true, uri, profileId);

  var cookie = cookie2 === null ? "" : cookie2;
  return cookie; // send cookie value to content
}


function documentCookieSetter(obj, contentDoc) {
  var profileId = new WindowProperties(contentDoc.defaultView).identity;

  switch (profileId) {
    case Profile.UnknownIdentity:
      return;
    case Profile.DefaultIdentity:
      util2.throwStack.go("documentCookieSetter " + profileId);
      return;
  }

  var originalUri = util2.stringToUri(contentDoc.location.href);
  Cookies.setCookie(profileId, originalUri, obj.value, true);
}


const PREF_COOKIE_BEHAVIOR = "network.cookie.cookieBehavior";

const Cookies = {
  _service: null,
  _prefs: null,

  start: function() {
    this._service = Cc["@mozilla.org/cookieService;1"].getService().QueryInterface(Ci.nsICookieService);
    this._prefs = Cc["@mozilla.org/preferences-service;1"].getService(Ci.nsIPrefBranch2);
    this._prefListener.behavior = this._prefs.getIntPref(PREF_COOKIE_BEHAVIOR);
    this._prefs.addObserver(PREF_COOKIE_BEHAVIOR, this._prefListener, false);
  },

  stop: function() {
    this._service = null;
    this._prefs.removeObserver(PREF_COOKIE_BEHAVIOR, this._prefListener);
    this._prefs = null;
  },


  _prefListener: {
    behavior: -1,

    QueryInterface: XPCOMUtils.generateQI([Ci.nsIObserver]),
    // aTopic=nsPref:changed aData=network.cookie.cookieBehavior
    observe: function(aSubject, aTopic, aData) {
      this.behavior = aSubject
                      .QueryInterface(Ci.nsIPrefBranch)
                      .getIntPref(PREF_COOKIE_BEHAVIOR);
      util.log("pref! " + aSubject + aTopic + aData + this.behavior);
    }

  },

  setCookie: function(profileId, originalUri, originalCookie, fromJs) {
    var uri = toInternalUri(originalUri, profileId);
    var val = convertCookieDomain(originalCookie, profileId);

    if (this._prefListener.behavior === 0) {
      this._setCookie(fromJs, uri, val);
      return;
    }

    var p = this._prefs;
    p.removeObserver(PREF_COOKIE_BEHAVIOR, this._prefListener);
    p.setIntPref(PREF_COOKIE_BEHAVIOR, 0);
    this._setCookie(fromJs, uri, val);
    p.setIntPref(PREF_COOKIE_BEHAVIOR, this._prefListener.behavior);
    p.addObserver(PREF_COOKIE_BEHAVIOR, this._prefListener, false);
  },

  getCookie: function(fromJs, originalUri, profileId) {
    var uri = toInternalUri(originalUri, profileId);

    if (this._prefListener.behavior === 0) {
      return this._getCookie(fromJs, uri);
    }

    var p = this._prefs;
    p.removeObserver(PREF_COOKIE_BEHAVIOR, this._prefListener);
    p.setIntPref(PREF_COOKIE_BEHAVIOR, 0);
    var cookie = this._getCookie(fromJs, uri);
    p.setIntPref(PREF_COOKIE_BEHAVIOR, this._prefListener.behavior);
    p.addObserver(PREF_COOKIE_BEHAVIOR, this._prefListener, false);
    return cookie;
  },

  _setCookie: function(fromJs, uri, val) {
    if (fromJs) {
      this._service.setCookieString(uri,
                                    null,
                                    val,
                                    null);
    } else {
      //setCookieString doesn't work for httponly cookies
      this._service.setCookieStringFromHttp(uri,   // aURI
                                            null,  // aFirstURI
                                            null,  // aPrompt
                                            val,   // aCookie
                                            null,  // aServerTime
                                            null); // aChannel
    }
  },

  _getCookie: function(fromJs, uri) {
    if (fromJs) {
      return this._service.getCookieString(uri,
                                           null);
    } else {
      return this._service.getCookieStringFromHttp(uri,   // aURI
                                                   null,  // aFirstURI
                                                   null); // aChannel
    }
  }
};


function convertCookieDomain(cookieHeader, profileId) {
  var objCookies = new SetCookieParser(cookieHeader, true);
  var len = objCookies.length;
  var newCookies = new Array(len);

  for (var idx = 0; idx < len; idx++) {
    var myCookie = objCookies.getCookieByIndex(idx);
    var realDomain = myCookie.getStringProperty("domain");
    if (realDomain.length > 0) {
      myCookie.defineMeta("domain", cookieInternalDomain(realDomain, profileId));
      newCookies[idx] = myCookie.toHeaderLine();
    } else {
      newCookies[idx] = myCookie.raw;
    }
  }

  return newCookies.join("\n");//objCookies.toHeader();
}


function toInternalUri(uri, sessionId) {
  var u = uri.clone();
  if (sessionId > Profile.DefaultIdentity) {
    u.host = cookieInternalDomain(u.host, sessionId);
  } else {
    util2.throwStack.go("invalid profile " + sessionId);
  }
  return u;
}


function cookieInternalDomain(domain, id) {
  return domain + ".multifox-profile-" + id;
}



function toLines(txt) {
  return txt.split(/\r\n|\r|\n/);
}


function SetCookieParser(cookieHeader, isSetCookie) { // TODO remove isSetCookie
  this.m_hasMeta = isSetCookie;
  this._allCookies = [];
  if (cookieHeader !== null) {
    var lines = toLines(cookieHeader);
    var len = lines.length;
    if (isSetCookie) {
      for (var idx = 0; idx < len; idx++) {
        this.parseLineSetCookie(lines[idx]);
      }
    } else {
      for (var idx = 0; idx < len; idx++) {
        this.parseLineCookie(lines[idx]);
      }
    }
  }
}

SetCookieParser.prototype = {
  parseLineSetCookie: function(headerLine) {
    var unit = new CookieUnit(headerLine);
    var items = headerLine.split(";");

    for (var idx = 0, len = items.length; idx < len; idx++) {
      var pair = splitValueName(items[idx]);
      var name = pair[0];
      var value = pair[1]; // null ==> name=HttpOnly, secure etc

      if (idx === 0) {
        if (name.length > 0 && value !== null) {
          unit.defineValue(name, value);
        } else {
          util2.logEx("_allCookies invalid", name, value, headerLine);
          break;
        }
      } else {
        unit.defineMeta(name, value);
      }
    }

    this._allCookies.push(unit);
  },

  parseLineCookie: function(headerLine) {
    var items = headerLine.split(";");
    for (var idx = 0, len = items.length; idx < len; idx++) {
      var unit = CookieUnit(items[idx]);
      var pair = splitValueName(items[idx]);
      unit.defineValue(pair[0], pair[1]);
      this._allCookies.push(unit);
    }
  },

  /*
  toHeader: function() {
    var allCookies = this._allCookies;
    var len = allCookies.length;
    //var buf = [];
    var buf = new Array(len);
    for (var idx = 0; idx < len; idx++) {
      //if (allCookies[idx].value !== null) {
      buf[idx] = allCookies[idx].toHeaderLine();
      //}
    }
    return this.m_hasMeta ? buf.join("\n") : buf.join(";");
  },

  getCookie: function(name) {
    var aCookie = this._allCookies;
    for (var idx = 0, len = aCookie.length; idx < len; idx++) {
      if (name === aCookie[idx].name) {
        return aCookie[idx];
      }
    }
    return null;
  },

  forEach: function(fn) {
    var c = this._allCookies;
    for (var idx = 0, len = c.length; idx < len; idx++) {
      fn(c[idx]);
    }
  }
  */

  get length() {
    return this._allCookies.length;
  },

  getCookieByIndex: function(idx) {
    return this._allCookies[idx];
  }
};

function CookieUnit(line) {
  this._data = {//instance value
    "_src": line
  };
}

CookieUnit.prototype = {
  _data: null,

  clone: function() {
    var c = new CookieUnit();
    for (var n in this._data) {
      c._data[n] = this._data[n];
    }
    return c;
  },

  get name() {
    var rv = this._data["_name"];
    return rv ? rv : "";
  },

  get value() {
    var rv = this._data["_value"];
    return rv ? rv : "";
  },

  get raw() {
    return this._data["_src"];
  },

  defineValue: function(name, val) {
    this._data["_name"] = name;
    this._data["_value"] = val;
  },

  //"secure":
  //"httponly":
  hasBooleanProperty: function(name) {
    //name = name.toLowerCase();
    return name in this._data;
  },

  setBooleanProperty: function(name, def) {
    if (def) {
      this._data[name] = null;
    } else {
      delete this._data[name];
    }
  },

  //"expires":
  //"domain":
  //"path":
  getStringProperty: function(name) {
    var rv = this._data[name];
    return rv ? rv : "";
    //return rv || "";
  },

  defineMeta: function(name, val) {
    name = name.toLowerCase();
    switch (name) {
      case "expires":
      case "domain":
      case "path":
      case "secure":
      case "httponly":
        this._data[name] = val;
        break;
    }
  },

  toHeaderLine: function() {//toString()
    var buf = [this.name + "=" + this.value];
    var props;

    props = ["secure", "httponly"];
    for (var idx = 0, len = props.length; idx < len; idx++) {
      var propName = props[idx];
      if (this.hasBooleanProperty(propName)) {
        buf.push(propName.toUpperCase());
      }
    }

    props = ["expires", "path", "domain"];
    for (var idx = 0, len = props.length; idx < len; idx++) {
      var propName = props[idx];
      var val = this.getStringProperty(propName);
      if (val.length > 0) {
        buf.push(propName.toUpperCase() + "=" + val);
      }
    }

    return buf.join(";");
  }
};


function splitValueName(cookie) {
  var idx = cookie.indexOf("=");
  if (idx === -1) {
    return [cookie.trim(), null];
  }

  // "a=bcd=e".split("=",2) returns [a,bcd]
  //   "abcde".split("=",2) returns [abcde]


  // MY =
  // 012^-----idx=3 length=4

  // MY =a:1=6
  // 012^-----idx=3 length=9

  var pair = ["", ""];
  pair[0] = cookie.substring(0, idx).trim();
  idx++;
  if (idx < cookie.length) {
    pair[1] = cookie.substring(idx);
  }

  return pair;
}
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2010
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */


function windowLocalStorage(obj, contentDoc) {
  var profileId = new WindowProperties(contentDoc.defaultView).identity;

  switch (profileId) {
    case Profile.UnknownIdentity:
      return;
    case Profile.DefaultIdentity:
      util2.throwStack.go("windowLocalStorage " + profileId);
      return;
  }


  var originalUri = util2.stringToUri(contentDoc.location.href);
  var uri = toInternalUri(originalUri, profileId);

  var principal = Cc["@mozilla.org/scriptsecuritymanager;1"]
                  .getService(Ci.nsIScriptSecurityManager)
                  .getCodebasePrincipal(uri);

  var storage = Cc["@mozilla.org/dom/storagemanager;1"]
                .getService(Ci.nsIDOMStorageManager)
                .getLocalStorageForPrincipal(principal);

  switch (obj.cmd) {
    case "clear":
      storage.clear();
      return;
    case "removeItem":
      storage.removeItem(obj.key);
      return;
    case "setItem":
      storage.setItem(obj.key, obj.val);
      return;
    case "getItem":
      rv = storage.getItem(obj.key);
      break;
    case "key":
      rv = storage.key(obj.index);
      break;
    case "length":
      rv = storage.length;
      break;
    default:
      throw obj.cmd;
  }

  return rv;
}
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */


const util2 = {

  browserWindowsEnum: function() {
    return Cc["@mozilla.org/appshell/window-mediator;1"]
            .getService(Ci.nsIWindowMediator)
            .getEnumerator("navigator:browser");
  },

  stringToUri: function(spec) {
    var io = Cc["@mozilla.org/network/io-service;1"].getService(Ci.nsIIOService);
    try {
      return io.newURI(spec, null, null);
    } catch (ex) {
      return null;
    }
  },

  logEx: function() {
    var buf = "logEx";// + this.logEx.caller + "\n";// + util2.throwStack._toString(false, Components.stack) + "===";
    for (var idx = 0, len = arguments.length; idx < len; idx++) {
      buf += "\n" + arguments[idx];// + "=" + this.caller[arguments[idx]];
    }

    buf += "\n====\n" + util2.throwStack._toString(false, Components.stack);
    util.log(buf);
  },

  throwStack: {

    go: function(txt) {
      util.log("Stack: " + txt + "\n" + this._toString(false, Components.stack));
      throw txt;
    },

    _toArray: function(webOnly, theStack) {
      var allItems = [];
      if (!theStack) {
        return allItems;
      }

      function StackItem() {}
      StackItem.prototype = {
        lang: "?",
        filename: "?",
        lineNumber: -1,
        name: "?"
      };

      // aparentemente  um bug que ocorre as vezes, todas as propriedades sao undefined
      if (theStack.languageName == undefined) {
        var item = new StackItem();
        item.name = theStack.toString();
        allItems.push(item);
      }

      // myStack.caller  quase sempre null, mas algumas vezes  undefined...
      for (var myStack = theStack; myStack; myStack = myStack.caller) {
        if (webOnly) {
          var n = myStack.filename;
          if (n === null) {
            continue;
          }

          if ((n.indexOf("http:") !== 0) && (n.indexOf("https:") !== 0)) {
            continue;
          }
        }


        var item = new StackItem();
        allItems.push(item);
        item.lang = myStack.languageName;
        item.filename = myStack.filename;
        item.lineNumber = myStack.lineNumber; // Valid line numbers begin at "1". "0" indicates unknown.
        item.name = myStack.name;
      }

      allItems.reverse();
      return allItems;
    },

    _toString: function(webOnly, theStack) {
      var arr = this._toArray(webOnly, theStack);
      var lines = new Array(arr.length);
      for (var idx = arr.length - 1, idx2 = 0; idx > -1; idx--) {
        var lang = webOnly ? "" : arr[idx].lang + " ";
        lines[idx2] = "[" + (idx + 1) + "] "
                    + lang
                    + arr[idx].filename + " "
                    + arr[idx].name + " ("
                    + arr[idx].lineNumber + ")";
        idx2++;
      }
      return lines.join("\n");
    }

  }
};
